@@ -6,6 +6,7 @@
 #define V8_OBJECTS_DATA_HANDLER_INL_H_
 
 #include "src/objects/data-handler.h"
+#include "src/objects/objects-inl.h"  // Needed for write barriers
 
 // Has to be the last include (doesn't have include guards):
 #include "src/objects/object-macros.h"
@@ -13,9 +14,7 @@
 namespace v8 {
 namespace internal {
 
-bool HeapObject::IsDataHandler() const {
-  return IsLoadHandler() || IsStoreHandler();
-}
+OBJECT_CONSTRUCTORS_IMPL(DataHandler, Struct)
 
 CAST_ACCESSOR(DataHandler)
 
@@ -23,15 +22,15 @@ ACCESSORS(DataHandler, smi_handler, Object, kSmiHandlerOffset)
 ACCESSORS(DataHandler, validity_cell, Object, kValidityCellOffset)
 
 int DataHandler::data_field_count() const {
-  return (map()->instance_size() - kSizeWithData0) / kPointerSize;
+  return (map().instance_size() - kSizeWithData0) / kTaggedSize;
 }
 
-ACCESSORS_CHECKED(DataHandler, data1, Object, kData1Offset,
-                  map()->instance_size() >= kSizeWithData1)
-ACCESSORS_CHECKED(DataHandler, data2, Object, kData2Offset,
-                  map()->instance_size() >= kSizeWithData2)
-ACCESSORS_CHECKED(DataHandler, data3, Object, kData3Offset,
-                  map()->instance_size() >= kSizeWithData3)
+WEAK_ACCESSORS_CHECKED(DataHandler, data1, kData1Offset,
+                       map().instance_size() >= kSizeWithData1)
+WEAK_ACCESSORS_CHECKED(DataHandler, data2, kData2Offset,
+                       map().instance_size() >= kSizeWithData2)
+WEAK_ACCESSORS_CHECKED(DataHandler, data3, kData3Offset,
+                       map().instance_size() >= kSizeWithData3)
 
 }  // namespace internal
 }  // namespace v8
