@@ -1,6 +1,12 @@
 'use strict';
 
-const { Math } = primordials;
+const {
+  BigInt,
+  Float32Array,
+  MathFloor,
+  Number,
+  Uint8Array,
+} = primordials;
 
 const {
   ERR_BUFFER_OUT_OF_BOUNDS,
@@ -22,6 +28,10 @@ const {
   ucs2Write,
   utf8Write
 } = internalBinding('buffer');
+const {
+  untransferable_object_private_symbol,
+  setHiddenValue,
+} = internalBinding('util');
 
 // Temporary buffers to convert numbers.
 const float32Array = new Float32Array(1);
@@ -43,13 +53,25 @@ function checkBounds(buf, offset, byteLength) {
 
 function checkInt(value, min, max, buf, offset, byteLength) {
   if (value > max || value < min) {
-    throw new ERR_OUT_OF_RANGE('value', `>= ${min} and <= ${max}`, value);
+    const n = typeof min === 'bigint' ? 'n' : '';
+    let range;
+    if (byteLength > 3) {
+      if (min === 0 || min === 0n) {
+        range = `>= 0${n} and < 2${n} ** ${(byteLength + 1) * 8}${n}`;
+      } else {
+        range = `>= -(2${n} ** ${(byteLength + 1) * 8 - 1}${n}) and < 2 ** ` +
+                `${(byteLength + 1) * 8 - 1}${n}`;
+      }
+    } else {
+      range = `>= ${min}${n} and <= ${max}${n}`;
+    }
+    throw new ERR_OUT_OF_RANGE('value', range, value);
   }
   checkBounds(buf, offset, byteLength);
 }
 
 function boundsError(value, length, type) {
-  if (Math.floor(value) !== value) {
+  if (MathFloor(value) !== value) {
     validateNumber(value, type);
     throw new ERR_OUT_OF_RANGE(type || 'offset', 'an integer', value);
   }
@@ -632,7 +654,7 @@ function writeU_Int48LE(buf, value, offset, min, max) {
   value = +value;
   checkInt(value, min, max, buf, offset, 5);
 
-  const newVal = Math.floor(value * 2 ** -32);
+  const newVal = MathFloor(value * 2 ** -32);
   buf[offset++] = value;
   value = value >>> 8;
   buf[offset++] = value;
@@ -657,7 +679,7 @@ function writeU_Int40LE(buf, value, offset, min, max) {
   buf[offset++] = value;
   value = value >>> 8;
   buf[offset++] = value;
-  buf[offset++] = Math.floor(newVal * 2 ** -32);
+  buf[offset++] = MathFloor(newVal * 2 ** -32);
   return offset;
 }
 
@@ -724,7 +746,7 @@ function writeUInt8(value, offset = 0) {
 
 function writeUIntBE(value, offset, byteLength) {
   if (byteLength === 6)
-    return writeU_Int48BE(this, value, offset, 0, 0xffffffffffffff);
+    return writeU_Int48BE(this, value, offset, 0, 0xffffffffffff);
   if (byteLength === 5)
     return writeU_Int40BE(this, value, offset, 0, 0xffffffffff);
   if (byteLength === 3)
@@ -743,7 +765,7 @@ function writeU_Int48BE(buf, value, offset, min, max) {
   value = +value;
   checkInt(value, min, max, buf, offset, 5);
 
-  const newVal = Math.floor(value * 2 ** -32);
+  const newVal = MathFloor(value * 2 ** -32);
   buf[offset++] = (newVal >>> 8);
   buf[offset++] = newVal;
   buf[offset + 3] = value;
@@ -760,7 +782,7 @@ function writeU_Int40BE(buf, value, offset, min, max) {
   value = +value;
   checkInt(value, min, max, buf, offset, 4);
 
-  buf[offset++] = Math.floor(value * 2 ** -32);
+  buf[offset++] = MathFloor(value * 2 ** -32);
   buf[offset + 3] = value;
   value = value >>> 8;
   buf[offset + 2] = value;
@@ -930,10 +952,14 @@ class FastBuffer extends Uint8Array {}
 function addBufferPrototypeMethods(proto) {
   proto.readBigUInt64LE = readBigUInt64LE,
   proto.readBigUInt64BE = readBigUInt64BE,
+  proto.readBigUint64LE = readBigUInt64LE,
+  proto.readBigUint64BE = readBigUInt64BE,
   proto.readBigInt64LE = readBigInt64LE,
   proto.readBigInt64BE = readBigInt64BE,
   proto.writeBigUInt64LE = writeBigUInt64LE,
   proto.writeBigUInt64BE = writeBigUInt64BE,
+  proto.writeBigUint64LE = writeBigUInt64LE,
+  proto.writeBigUint64BE = writeBigUInt64BE,
   proto.writeBigInt64LE = writeBigInt64LE,
   proto.writeBigInt64BE = writeBigInt64BE,
 
@@ -944,6 +970,13 @@ function addBufferPrototypeMethods(proto) {
   proto.readUIntBE = readUIntBE;
   proto.readUInt32BE = readUInt32BE;
   proto.readUInt16BE = readUInt16BE;
+  proto.readUintLE = readUIntLE;
+  proto.readUint32LE = readUInt32LE;
+  proto.readUint16LE = readUInt16LE;
+  proto.readUint8 = readUInt8;
+  proto.readUintBE = readUIntBE;
+  proto.readUint32BE = readUInt32BE;
+  proto.readUint16BE = readUInt16BE;
   proto.readIntLE = readIntLE;
   proto.readInt32LE = readInt32LE;
   proto.readInt16LE = readInt16LE;
@@ -959,6 +992,13 @@ function addBufferPrototypeMethods(proto) {
   proto.writeUIntBE = writeUIntBE;
   proto.writeUInt32BE = writeUInt32BE;
   proto.writeUInt16BE = writeUInt16BE;
+  proto.writeUintLE = writeUIntLE;
+  proto.writeUint32LE = writeUInt32LE;
+  proto.writeUint16LE = writeUInt16LE;
+  proto.writeUint8 = writeUInt8;
+  proto.writeUintBE = writeUIntBE;
+  proto.writeUint32BE = writeUInt32BE;
+  proto.writeUint16BE = writeUInt16BE;
   proto.writeIntLE = writeIntLE;
   proto.writeInt32LE = writeInt32LE;
   proto.writeInt16LE = writeInt16LE;
@@ -990,7 +1030,16 @@ function addBufferPrototypeMethods(proto) {
   proto.utf8Write = utf8Write;
 }
 
+// This would better be placed in internal/worker/io.js, but that doesn't work
+// because Buffer needs this and that would introduce a cyclic dependency.
+function markAsUntransferable(obj) {
+  if ((typeof obj !== 'object' && typeof obj !== 'function') || obj === null)
+    return;  // This object is a primitive and therefore already untransferable.
+  setHiddenValue(obj, untransferable_object_private_symbol, true);
+}
+
 module.exports = {
   FastBuffer,
-  addBufferPrototypeMethods
+  addBufferPrototypeMethods,
+  markAsUntransferable,
 };
