@@ -22,38 +22,66 @@
 'use strict';
 
 const {
-  ERR_BUFFER_TOO_LARGE,
-  ERR_INVALID_ARG_TYPE,
-  ERR_OUT_OF_RANGE,
-  ERR_ZLIB_INITIALIZATION_FAILED
-} = require('internal/errors').codes;
+  ArrayBuffer,
+  Error,
+  MathMax,
+  NumberIsFinite,
+  NumberIsNaN,
+  ObjectDefineProperties,
+  ObjectDefineProperty,
+  ObjectFreeze,
+  ObjectGetPrototypeOf,
+  ObjectKeys,
+  ObjectSetPrototypeOf,
+  Symbol,
+  Uint32Array,
+} = primordials;
+
+const {
+  codes: {
+    ERR_BROTLI_INVALID_PARAM,
+    ERR_BUFFER_TOO_LARGE,
+    ERR_INVALID_ARG_TYPE,
+    ERR_OUT_OF_RANGE,
+    ERR_ZLIB_INITIALIZATION_FAILED,
+  },
+  hideStackFrames
+} = require('internal/errors');
 const Transform = require('_stream_transform');
 const {
-  _extend,
-  inherits,
-  types: {
-    isAnyArrayBuffer,
-    isArrayBufferView
-  }
-} = require('util');
-const binding = process.binding('zlib');
-const assert = require('assert').ok;
+  deprecate
+} = require('internal/util');
+const {
+  isArrayBufferView,
+  isAnyArrayBuffer
+} = require('internal/util/types');
+const binding = internalBinding('zlib');
+const assert = require('internal/assert');
 const {
   Buffer,
   kMaxLength
 } = require('buffer');
 const { owner_symbol } = require('internal/async_hooks').symbols;
 
-const constants = process.binding('constants').zlib;
+const kFlushFlag = Symbol('kFlushFlag');
+
+const constants = internalBinding('constants').zlib;
 const {
+  // Zlib flush levels
   Z_NO_FLUSH, Z_BLOCK, Z_PARTIAL_FLUSH, Z_SYNC_FLUSH, Z_FULL_FLUSH, Z_FINISH,
+  // Zlib option values
   Z_MIN_CHUNK, Z_MIN_WINDOWBITS, Z_MAX_WINDOWBITS, Z_MIN_LEVEL, Z_MAX_LEVEL,
   Z_MIN_MEMLEVEL, Z_MAX_MEMLEVEL, Z_DEFAULT_CHUNK, Z_DEFAULT_COMPRESSION,
   Z_DEFAULT_STRATEGY, Z_DEFAULT_WINDOWBITS, Z_DEFAULT_MEMLEVEL, Z_FIXED,
-  DEFLATE, DEFLATERAW, INFLATE, INFLATERAW, GZIP, GUNZIP, UNZIP
+  // Node's compression stream modes (node_zlib_mode)
+  DEFLATE, DEFLATERAW, INFLATE, INFLATERAW, GZIP, GUNZIP, UNZIP,
+  BROTLI_DECODE, BROTLI_ENCODE,
+  // Brotli operations (~flush levels)
+  BROTLI_OPERATION_PROCESS, BROTLI_OPERATION_FLUSH,
+  BROTLI_OPERATION_FINISH
 } = constants;
 
-// translation table for return codes.
+// Translation table for return codes.
 const codes = {
   Z_OK: constants.Z_OK,
   Z_STREAM_END: constants.Z_STREAM_END,
@@ -66,17 +94,17 @@ const codes = {
   Z_VERSION_ERROR: constants.Z_VERSION_ERROR
 };
 
-const ckeys = Object.keys(codes);
-for (var ck = 0; ck < ckeys.length; ck++) {
-  var ckey = ckeys[ck];
+for (const ckey of ObjectKeys(codes)) {
   codes[codes[ckey]] = ckey;
 }
 
 function zlibBuffer(engine, buffer, callback) {
+  if (typeof callback !== 'function')
+    throw new ERR_INVALID_ARG_TYPE('callback', 'function', callback);
   // Streams do not support non-Buffer ArrayBufferViews yet. Convert it to a
   // Buffer without copying.
   if (isArrayBufferView(buffer) &&
-      Object.getPrototypeOf(buffer) !== Buffer.prototype) {
+      ObjectGetPrototypeOf(buffer) !== Buffer.prototype) {
     buffer = Buffer.from(buffer.buffer, buffer.byteOffset, buffer.byteLength);
   } else if (isAnyArrayBuffer(buffer)) {
     buffer = Buffer.from(buffer);
@@ -96,6 +124,11 @@ function zlibBufferOnData(chunk) {
   else
     this.buffers.push(chunk);
   this.nread += chunk.length;
+  if (this.nread > this._maxOutputLength) {
+    this.close();
+    this.removeAllListeners('end');
+    this.cb(new ERR_BUFFER_TOO_LARGE(this._maxOutputLength));
+  }
 }
 
 function zlibBufferOnError(err) {
@@ -104,20 +137,15 @@ function zlibBufferOnError(err) {
 }
 
 function zlibBufferOnEnd() {
-  var buf;
-  var err;
-  if (this.nread >= kMaxLength) {
-    err = new ERR_BUFFER_TOO_LARGE();
-  } else if (this.nread === 0) {
+  let buf;
+  if (this.nread === 0) {
     buf = Buffer.alloc(0);
   } else {
-    var bufs = this.buffers;
+    const bufs = this.buffers;
     buf = (bufs.length === 1 ? bufs[0] : Buffer.concat(bufs, this.nread));
   }
   this.close();
-  if (err)
-    this.cb(err);
-  else if (this._info)
+  if (this._info)
     this.cb(null, { buffer: buf, engine: this });
   else
     this.cb(null, buf);
@@ -144,9 +172,9 @@ function zlibBufferSync(engine, buffer) {
 }
 
 function zlibOnError(message, errno, code) {
-  var self = this[owner_symbol];
-  // there is no way to cleanly recover.
-  // continuing only obscures problems.
+  const self = this[owner_symbol];
+  // There is no way to cleanly recover.
+  // Continuing only obscures problems.
   _close(self);
   self._hadError = true;
 
@@ -161,68 +189,54 @@ function zlibOnError(message, errno, code) {
 // 2. Returns true for finite numbers
 // 3. Throws ERR_INVALID_ARG_TYPE for non-numbers
 // 4. Throws ERR_OUT_OF_RANGE for infinite numbers
-function checkFiniteNumber(number, name) {
+const checkFiniteNumber = hideStackFrames((number, name) => {
   // Common case
   if (number === undefined) {
     return false;
   }
 
-  if (Number.isFinite(number)) {
+  if (NumberIsFinite(number)) {
     return true; // Is a valid number
   }
 
-  if (Number.isNaN(number)) {
+  if (NumberIsNaN(number)) {
     return false;
   }
 
   // Other non-numbers
   if (typeof number !== 'number') {
-    const err = new ERR_INVALID_ARG_TYPE(name, 'number', number);
-    Error.captureStackTrace(err, checkFiniteNumber);
-    throw err;
+    throw new ERR_INVALID_ARG_TYPE(name, 'number', number);
   }
 
   // Infinite numbers
-  const err = new ERR_OUT_OF_RANGE(name, 'a finite number', number);
-  Error.captureStackTrace(err, checkFiniteNumber);
-  throw err;
-}
+  throw new ERR_OUT_OF_RANGE(name, 'a finite number', number);
+});
 
 // 1. Returns def for number when it's undefined or NaN
 // 2. Returns number for finite numbers >= lower and <= upper
 // 3. Throws ERR_INVALID_ARG_TYPE for non-numbers
 // 4. Throws ERR_OUT_OF_RANGE for infinite numbers or numbers > upper or < lower
-function checkRangesOrGetDefault(number, name, lower, upper, def) {
-  if (!checkFiniteNumber(number, name)) {
-    return def;
+const checkRangesOrGetDefault = hideStackFrames(
+  (number, name, lower, upper, def) => {
+    if (!checkFiniteNumber(number, name)) {
+      return def;
+    }
+    if (number < lower || number > upper) {
+      throw new ERR_OUT_OF_RANGE(name,
+                                 `>= ${lower} and <= ${upper}`, number);
+    }
+    return number;
   }
-  if (number < lower || number > upper) {
-    const err = new ERR_OUT_OF_RANGE(name,
-                                     `>= ${lower} and <= ${upper}`, number);
-    Error.captureStackTrace(err, checkRangesOrGetDefault);
-    throw err;
-  }
-  return number;
-}
+);
 
-// the Zlib class they all inherit from
-// This thing manages the queue of requests, and returns
-// true or false if there is anything in the queue when
-// you call the .write() method.
-function Zlib(opts, mode) {
-  var chunkSize = Z_DEFAULT_CHUNK;
-  var flush = Z_NO_FLUSH;
-  var finishFlush = Z_FINISH;
-  var windowBits = Z_DEFAULT_WINDOWBITS;
-  var level = Z_DEFAULT_COMPRESSION;
-  var memLevel = Z_DEFAULT_MEMLEVEL;
-  var strategy = Z_DEFAULT_STRATEGY;
-  var dictionary;
-
-  // The Zlib class is not exported to user land, the mode should only be
+// The base class for all Zlib-style streams.
+function ZlibBase(opts, mode, handle, { flush, finishFlush, fullFlush }) {
+  let chunkSize = Z_DEFAULT_CHUNK;
+  let maxOutputLength = kMaxLength;
+  // The ZlibBase class is not exported to user land, the mode should only be
   // passed in by us.
   assert(typeof mode === 'number');
-  assert(mode >= DEFLATE && mode <= UNZIP);
+  assert(mode >= DEFLATE && mode <= BROTLI_ENCODE);
 
   if (opts) {
     chunkSize = opts.chunkSize;
@@ -235,12 +249,369 @@ function Zlib(opts, mode) {
 
     flush = checkRangesOrGetDefault(
       opts.flush, 'options.flush',
-      Z_NO_FLUSH, Z_BLOCK, Z_NO_FLUSH);
+      Z_NO_FLUSH, Z_BLOCK, flush);
 
     finishFlush = checkRangesOrGetDefault(
       opts.finishFlush, 'options.finishFlush',
-      Z_NO_FLUSH, Z_BLOCK, Z_FINISH);
+      Z_NO_FLUSH, Z_BLOCK, finishFlush);
 
+    maxOutputLength = checkRangesOrGetDefault(
+      opts.maxOutputLength, 'options.maxOutputLength',
+      1, kMaxLength, kMaxLength);
+
+    if (opts.encoding || opts.objectMode || opts.writableObjectMode) {
+      opts = { ...opts };
+      opts.encoding = null;
+      opts.objectMode = false;
+      opts.writableObjectMode = false;
+    }
+  }
+
+  Transform.call(this, opts);
+  this._hadError = false;
+  this.bytesWritten = 0;
+  this._handle = handle;
+  handle[owner_symbol] = this;
+  // Used by processCallback() and zlibOnError()
+  handle.onerror = zlibOnError;
+  this._outBuffer = Buffer.allocUnsafe(chunkSize);
+  this._outOffset = 0;
+
+  this._chunkSize = chunkSize;
+  this._defaultFlushFlag = flush;
+  this._finishFlushFlag = finishFlush;
+  this._defaultFullFlushFlag = fullFlush;
+  this.once('end', _close.bind(null, this));
+  this._info = opts && opts.info;
+  this._maxOutputLength = maxOutputLength;
+}
+ObjectSetPrototypeOf(ZlibBase.prototype, Transform.prototype);
+ObjectSetPrototypeOf(ZlibBase, Transform);
+
+ObjectDefineProperty(ZlibBase.prototype, '_closed', {
+  configurable: true,
+  enumerable: true,
+  get() {
+    return !this._handle;
+  }
+});
+
+// `bytesRead` made sense as a name when looking from the zlib engine's
+// perspective, but it is inconsistent with all other streams exposed by Node.js
+// that have this concept, where it stands for the number of bytes read
+// *from* the stream (that is, net.Socket/tls.Socket & file system streams).
+ObjectDefineProperty(ZlibBase.prototype, 'bytesRead', {
+  configurable: true,
+  enumerable: true,
+  get: deprecate(function() {
+    return this.bytesWritten;
+  }, 'zlib.bytesRead is deprecated and will change its meaning in the ' +
+     'future. Use zlib.bytesWritten instead.', 'DEP0108'),
+  set: deprecate(function(value) {
+    this.bytesWritten = value;
+  }, 'Setting zlib.bytesRead is deprecated. ' +
+     'This feature will be removed in the future.', 'DEP0108')
+});
+
+ZlibBase.prototype.reset = function() {
+  if (!this._handle)
+    assert(false, 'zlib binding closed');
+  return this._handle.reset();
+};
+
+// This is the _flush function called by the transform class,
+// internally, when the last chunk has been written.
+ZlibBase.prototype._flush = function(callback) {
+  this._transform(Buffer.alloc(0), '', callback);
+};
+
+// If a flush is scheduled while another flush is still pending, a way to figure
+// out which one is the "stronger" flush is needed.
+// This is currently only used to figure out which flush flag to use for the
+// last chunk.
+// Roughly, the following holds:
+// Z_NO_FLUSH (< Z_TREES) < Z_BLOCK < Z_PARTIAL_FLUSH <
+//     Z_SYNC_FLUSH < Z_FULL_FLUSH < Z_FINISH
+const flushiness = [];
+let i = 0;
+const kFlushFlagList = [Z_NO_FLUSH, Z_BLOCK, Z_PARTIAL_FLUSH,
+                        Z_SYNC_FLUSH, Z_FULL_FLUSH, Z_FINISH];
+for (const flushFlag of kFlushFlagList) {
+  flushiness[flushFlag] = i++;
+}
+
+function maxFlush(a, b) {
+  return flushiness[a] > flushiness[b] ? a : b;
+}
+
+// Set up a list of 'special' buffers that can be written using .write()
+// from the .flush() code as a way of introducing flushing operations into the
+// write sequence.
+const kFlushBuffers = [];
+{
+  const dummyArrayBuffer = new ArrayBuffer();
+  for (const flushFlag of kFlushFlagList) {
+    kFlushBuffers[flushFlag] = Buffer.from(dummyArrayBuffer);
+    kFlushBuffers[flushFlag][kFlushFlag] = flushFlag;
+  }
+}
+
+ZlibBase.prototype.flush = function(kind, callback) {
+  if (typeof kind === 'function' || (kind === undefined && !callback)) {
+    callback = kind;
+    kind = this._defaultFullFlushFlag;
+  }
+
+  if (this.writableFinished) {
+    if (callback)
+      process.nextTick(callback);
+  } else if (this.writableEnded) {
+    if (callback)
+      this.once('end', callback);
+  } else {
+    this.write(kFlushBuffers[kind], '', callback);
+  }
+};
+
+ZlibBase.prototype.close = function(callback) {
+  _close(this, callback);
+  this.destroy();
+};
+
+ZlibBase.prototype._destroy = function(err, callback) {
+  _close(this);
+  callback(err);
+};
+
+ZlibBase.prototype._transform = function(chunk, encoding, cb) {
+  let flushFlag = this._defaultFlushFlag;
+  // We use a 'fake' zero-length chunk to carry information about flushes from
+  // the public API to the actual stream implementation.
+  if (typeof chunk[kFlushFlag] === 'number') {
+    flushFlag = chunk[kFlushFlag];
+  }
+
+  // For the last chunk, also apply `_finishFlushFlag`.
+  if (this.writableEnded && this.writableLength === chunk.byteLength) {
+    flushFlag = maxFlush(flushFlag, this._finishFlushFlag);
+  }
+  processChunk(this, chunk, flushFlag, cb);
+};
+
+ZlibBase.prototype._processChunk = function(chunk, flushFlag, cb) {
+  // _processChunk() is left for backwards compatibility
+  if (typeof cb === 'function')
+    processChunk(this, chunk, flushFlag, cb);
+  else
+    return processChunkSync(this, chunk, flushFlag);
+};
+
+function processChunkSync(self, chunk, flushFlag) {
+  let availInBefore = chunk.byteLength;
+  let availOutBefore = self._chunkSize - self._outOffset;
+  let inOff = 0;
+  let availOutAfter;
+  let availInAfter;
+
+  let buffers = null;
+  let nread = 0;
+  let inputRead = 0;
+  const state = self._writeState;
+  const handle = self._handle;
+  let buffer = self._outBuffer;
+  let offset = self._outOffset;
+  const chunkSize = self._chunkSize;
+
+  let error;
+  self.on('error', function onError(er) {
+    error = er;
+  });
+
+  while (true) {
+    handle.writeSync(flushFlag,
+                     chunk, // in
+                     inOff, // in_off
+                     availInBefore, // in_len
+                     buffer, // out
+                     offset, // out_off
+                     availOutBefore); // out_len
+    if (error)
+      throw error;
+
+    availOutAfter = state[0];
+    availInAfter = state[1];
+
+    const inDelta = (availInBefore - availInAfter);
+    inputRead += inDelta;
+
+    const have = availOutBefore - availOutAfter;
+    if (have > 0) {
+      const out = buffer.slice(offset, offset + have);
+      offset += have;
+      if (!buffers)
+        buffers = [out];
+      else
+        buffers.push(out);
+      nread += out.byteLength;
+
+      if (nread > self._maxOutputLength) {
+        _close(self);
+        throw new ERR_BUFFER_TOO_LARGE(self._maxOutputLength);
+      }
+
+    } else {
+      assert(have === 0, 'have should not go down');
+    }
+
+    // Exhausted the output buffer, or used all the input create a new one.
+    if (availOutAfter === 0 || offset >= chunkSize) {
+      availOutBefore = chunkSize;
+      offset = 0;
+      buffer = Buffer.allocUnsafe(chunkSize);
+    }
+
+    if (availOutAfter === 0) {
+      // Not actually done. Need to reprocess.
+      // Also, update the availInBefore to the availInAfter value,
+      // so that if we have to hit it a third (fourth, etc.) time,
+      // it'll have the correct byte counts.
+      inOff += inDelta;
+      availInBefore = availInAfter;
+    } else {
+      break;
+    }
+  }
+
+  self.bytesWritten = inputRead;
+  _close(self);
+
+  if (nread === 0)
+    return Buffer.alloc(0);
+
+  return (buffers.length === 1 ? buffers[0] : Buffer.concat(buffers, nread));
+}
+
+function processChunk(self, chunk, flushFlag, cb) {
+  const handle = self._handle;
+  if (!handle) return process.nextTick(cb);
+
+  handle.buffer = chunk;
+  handle.cb = cb;
+  handle.availOutBefore = self._chunkSize - self._outOffset;
+  handle.availInBefore = chunk.byteLength;
+  handle.inOff = 0;
+  handle.flushFlag = flushFlag;
+
+  handle.write(flushFlag,
+               chunk, // in
+               0, // in_off
+               handle.availInBefore, // in_len
+               self._outBuffer, // out
+               self._outOffset, // out_off
+               handle.availOutBefore); // out_len
+}
+
+function processCallback() {
+  // This callback's context (`this`) is the `_handle` (ZCtx) object. It is
+  // important to null out the values once they are no longer needed since
+  // `_handle` can stay in memory long after the buffer is needed.
+  const handle = this;
+  const self = this[owner_symbol];
+  const state = self._writeState;
+
+  if (self._hadError || self.destroyed) {
+    this.buffer = null;
+    this.cb();
+    return;
+  }
+
+  const availOutAfter = state[0];
+  const availInAfter = state[1];
+
+  const inDelta = handle.availInBefore - availInAfter;
+  self.bytesWritten += inDelta;
+
+  const have = handle.availOutBefore - availOutAfter;
+  if (have > 0) {
+    const out = self._outBuffer.slice(self._outOffset, self._outOffset + have);
+    self._outOffset += have;
+    self.push(out);
+  } else {
+    assert(have === 0, 'have should not go down');
+  }
+
+  if (self.destroyed) {
+    this.cb();
+    return;
+  }
+
+  // Exhausted the output buffer, or used all the input create a new one.
+  if (availOutAfter === 0 || self._outOffset >= self._chunkSize) {
+    handle.availOutBefore = self._chunkSize;
+    self._outOffset = 0;
+    self._outBuffer = Buffer.allocUnsafe(self._chunkSize);
+  }
+
+  if (availOutAfter === 0) {
+    // Not actually done. Need to reprocess.
+    // Also, update the availInBefore to the availInAfter value,
+    // so that if we have to hit it a third (fourth, etc.) time,
+    // it'll have the correct byte counts.
+    handle.inOff += inDelta;
+    handle.availInBefore = availInAfter;
+
+    this.write(handle.flushFlag,
+               this.buffer, // in
+               handle.inOff, // in_off
+               handle.availInBefore, // in_len
+               self._outBuffer, // out
+               self._outOffset, // out_off
+               self._chunkSize); // out_len
+    return;
+  }
+
+  if (availInAfter > 0) {
+    // If we have more input that should be written, but we also have output
+    // space available, that means that the compression library was not
+    // interested in receiving more data, and in particular that the input
+    // stream has ended early.
+    // This applies to streams where we don't check data past the end of
+    // what was consumed; that is, everything except Gunzip/Unzip.
+    self.push(null);
+  }
+
+  // Finished with the chunk.
+  this.buffer = null;
+  this.cb();
+}
+
+function _close(engine, callback) {
+  if (callback)
+    process.nextTick(callback);
+
+  // Caller may invoke .close after a zlib error (which will null _handle).
+  if (!engine._handle)
+    return;
+
+  engine._handle.close();
+  engine._handle = null;
+}
+
+const zlibDefaultOpts = {
+  flush: Z_NO_FLUSH,
+  finishFlush: Z_FINISH,
+  fullFlush: Z_FULL_FLUSH
+};
+// Base class for all streams actually backed by zlib and using zlib-specific
+// parameters.
+function Zlib(opts, mode) {
+  let windowBits = Z_DEFAULT_WINDOWBITS;
+  let level = Z_DEFAULT_COMPRESSION;
+  let memLevel = Z_DEFAULT_MEMLEVEL;
+  let strategy = Z_DEFAULT_STRATEGY;
+  let dictionary;
+
+  if (opts) {
     // windowBits is special. On the compression side, 0 is an invalid value.
     // But on the decompression side, a value of 0 for windowBits tells zlib
     // to use the window size in the zlib header of the compressed stream.
@@ -250,9 +621,11 @@ function Zlib(opts, mode) {
          mode === UNZIP)) {
       windowBits = 0;
     } else {
+      // `{ windowBits: 8 }` is valid for deflate but not gzip.
+      const min = Z_MIN_WINDOWBITS + (mode === GZIP ? 1 : 0);
       windowBits = checkRangesOrGetDefault(
         opts.windowBits, 'options.windowBits',
-        Z_MIN_WINDOWBITS, Z_MAX_WINDOWBITS, Z_DEFAULT_WINDOWBITS);
+        min, Z_MAX_WINDOWBITS, Z_DEFAULT_WINDOWBITS);
     }
 
     level = checkRangesOrGetDefault(
@@ -279,68 +652,28 @@ function Zlib(opts, mode) {
         );
       }
     }
-
-    if (opts.encoding || opts.objectMode || opts.writableObjectMode) {
-      opts = _extend({}, opts);
-      opts.encoding = null;
-      opts.objectMode = false;
-      opts.writableObjectMode = false;
-    }
   }
-  Transform.call(this, opts);
-  this.bytesWritten = 0;
-  this._handle = new binding.Zlib(mode);
-  // Used by processCallback() and zlibOnError()
-  this._handle[owner_symbol] = this;
-  this._handle.onerror = zlibOnError;
-  this._hadError = false;
+
+  const handle = new binding.Zlib(mode);
+  // Ideally, we could let ZlibBase() set up _writeState. I haven't been able
+  // to come up with a good solution that doesn't break our internal API,
+  // and with it all supported npm versions at the time of writing.
   this._writeState = new Uint32Array(2);
+  handle.init(windowBits,
+              level,
+              memLevel,
+              strategy,
+              this._writeState,
+              processCallback,
+              dictionary);
 
-  if (!this._handle.init(windowBits,
-                         level,
-                         memLevel,
-                         strategy,
-                         this._writeState,
-                         processCallback,
-                         dictionary)) {
-    throw new ERR_ZLIB_INITIALIZATION_FAILED();
-  }
+  ZlibBase.call(this, opts, mode, handle, zlibDefaultOpts);
 
-  this._outBuffer = Buffer.allocUnsafe(chunkSize);
-  this._outOffset = 0;
   this._level = level;
   this._strategy = strategy;
-  this._chunkSize = chunkSize;
-  this._defaultFlushFlag = flush;
-  this._finishFlushFlag = finishFlush;
-  this._nextFlush = -1;
-  this._info = opts && opts.info;
-  this.once('end', this.close);
 }
-inherits(Zlib, Transform);
-
-Object.defineProperty(Zlib.prototype, '_closed', {
-  configurable: true,
-  enumerable: true,
-  get() {
-    return !this._handle;
-  }
-});
-
-// `bytesRead` made sense as a name when looking from the zlib engine's
-// perspective, but it is inconsistent with all other streams exposed by Node.js
-// that have this concept, where it stands for the number of bytes read
-// *from* the stream (that is, net.Socket/tls.Socket & file system streams).
-Object.defineProperty(Zlib.prototype, 'bytesRead', {
-  configurable: true,
-  enumerable: true,
-  get() {
-    return this.bytesWritten;
-  },
-  set(value) {
-    this.bytesWritten = value;
-  }
-});
+ObjectSetPrototypeOf(Zlib.prototype, ZlibBase.prototype);
+ObjectSetPrototypeOf(Zlib, ZlibBase);
 
 // This callback is used by `.params()` to wait until a full flush happened
 // before adjusting the parameters. In particular, the call to the native
@@ -368,276 +701,6 @@ Zlib.prototype.params = function params(level, strategy, callback) {
   }
 };
 
-Zlib.prototype.reset = function reset() {
-  if (!this._handle)
-    assert(false, 'zlib binding closed');
-  return this._handle.reset();
-};
-
-// This is the _flush function called by the transform class,
-// internally, when the last chunk has been written.
-Zlib.prototype._flush = function _flush(callback) {
-  this._transform(Buffer.alloc(0), '', callback);
-};
-
-// If a flush is scheduled while another flush is still pending, a way to figure
-// out which one is the "stronger" flush is needed.
-// Roughly, the following holds:
-// Z_NO_FLUSH (< Z_TREES) < Z_BLOCK < Z_PARTIAL_FLUSH <
-//     Z_SYNC_FLUSH < Z_FULL_FLUSH < Z_FINISH
-const flushiness = [];
-let i = 0;
-for (const flushFlag of [Z_NO_FLUSH, Z_BLOCK, Z_PARTIAL_FLUSH,
-                         Z_SYNC_FLUSH, Z_FULL_FLUSH, Z_FINISH]) {
-  flushiness[flushFlag] = i++;
-}
-
-function maxFlush(a, b) {
-  return flushiness[a] > flushiness[b] ? a : b;
-}
-
-const flushBuffer = Buffer.alloc(0);
-Zlib.prototype.flush = function flush(kind, callback) {
-  var ws = this._writableState;
-
-  if (typeof kind === 'function' || (kind === undefined && !callback)) {
-    callback = kind;
-    kind = Z_FULL_FLUSH;
-  }
-
-  if (ws.ended) {
-    if (callback)
-      process.nextTick(callback);
-  } else if (ws.ending) {
-    if (callback)
-      this.once('end', callback);
-  } else if (this._nextFlush !== -1) {
-    // This means that there is a flush currently in the write queue.
-    // We currently coalesce this flush into the pending one.
-    this._nextFlush = maxFlush(this._nextFlush, kind);
-  } else {
-    this._nextFlush = kind;
-    this.write(flushBuffer, '', callback);
-  }
-};
-
-Zlib.prototype.close = function close(callback) {
-  _close(this, callback);
-  this.destroy();
-};
-
-Zlib.prototype._destroy = function _destroy(err, callback) {
-  _close(this);
-  callback(err);
-};
-
-Zlib.prototype._transform = function _transform(chunk, encoding, cb) {
-  var flushFlag = this._defaultFlushFlag;
-  // We use a 'fake' zero-length chunk to carry information about flushes from
-  // the public API to the actual stream implementation.
-  if (chunk === flushBuffer) {
-    flushFlag = this._nextFlush;
-    this._nextFlush = -1;
-  }
-
-  // For the last chunk, also apply `_finishFlushFlag`.
-  var ws = this._writableState;
-  if ((ws.ending || ws.ended) && ws.length === chunk.byteLength) {
-    flushFlag = maxFlush(flushFlag, this._finishFlushFlag);
-  }
-  processChunk(this, chunk, flushFlag, cb);
-};
-
-Zlib.prototype._processChunk = function _processChunk(chunk, flushFlag, cb) {
-  // _processChunk() is left for backwards compatibility
-  if (typeof cb === 'function')
-    processChunk(this, chunk, flushFlag, cb);
-  else
-    return processChunkSync(this, chunk, flushFlag);
-};
-
-function processChunkSync(self, chunk, flushFlag) {
-  var availInBefore = chunk.byteLength;
-  var availOutBefore = self._chunkSize - self._outOffset;
-  var inOff = 0;
-  var availOutAfter;
-  var availInAfter;
-
-  var buffers = null;
-  var nread = 0;
-  var inputRead = 0;
-  var state = self._writeState;
-  var handle = self._handle;
-  var buffer = self._outBuffer;
-  var offset = self._outOffset;
-  var chunkSize = self._chunkSize;
-
-  var error;
-  self.on('error', function onError(er) {
-    error = er;
-  });
-
-  while (true) {
-    handle.writeSync(flushFlag,
-                     chunk, // in
-                     inOff, // in_off
-                     availInBefore, // in_len
-                     buffer, // out
-                     offset, // out_off
-                     availOutBefore); // out_len
-    if (error)
-      throw error;
-
-    availOutAfter = state[0];
-    availInAfter = state[1];
-
-    var inDelta = (availInBefore - availInAfter);
-    inputRead += inDelta;
-
-    var have = availOutBefore - availOutAfter;
-    if (have > 0) {
-      var out = buffer.slice(offset, offset + have);
-      offset += have;
-      if (!buffers)
-        buffers = [out];
-      else
-        buffers.push(out);
-      nread += out.byteLength;
-    } else {
-      assert(have === 0, 'have should not go down');
-    }
-
-    // exhausted the output buffer, or used all the input create a new one.
-    if (availOutAfter === 0 || offset >= chunkSize) {
-      availOutBefore = chunkSize;
-      offset = 0;
-      buffer = Buffer.allocUnsafe(chunkSize);
-    }
-
-    if (availOutAfter === 0) {
-      // Not actually done. Need to reprocess.
-      // Also, update the availInBefore to the availInAfter value,
-      // so that if we have to hit it a third (fourth, etc.) time,
-      // it'll have the correct byte counts.
-      inOff += inDelta;
-      availInBefore = availInAfter;
-    } else {
-      break;
-    }
-  }
-
-  self.bytesWritten = inputRead;
-  _close(self);
-
-  if (nread >= kMaxLength) {
-    throw new ERR_BUFFER_TOO_LARGE();
-  }
-
-  if (nread === 0)
-    return Buffer.alloc(0);
-
-  return (buffers.length === 1 ? buffers[0] : Buffer.concat(buffers, nread));
-}
-
-function processChunk(self, chunk, flushFlag, cb) {
-  var handle = self._handle;
-  assert(handle, 'zlib binding closed');
-
-  handle.buffer = chunk;
-  handle.cb = cb;
-  handle.availOutBefore = self._chunkSize - self._outOffset;
-  handle.availInBefore = chunk.byteLength;
-  handle.inOff = 0;
-  handle.flushFlag = flushFlag;
-
-  handle.write(flushFlag,
-               chunk, // in
-               0, // in_off
-               handle.availInBefore, // in_len
-               self._outBuffer, // out
-               self._outOffset, // out_off
-               handle.availOutBefore); // out_len
-}
-
-function processCallback() {
-  // This callback's context (`this`) is the `_handle` (ZCtx) object. It is
-  // important to null out the values once they are no longer needed since
-  // `_handle` can stay in memory long after the buffer is needed.
-  var handle = this;
-  var self = this.jsref;
-  var state = self._writeState;
-
-  if (self._hadError) {
-    this.buffer = null;
-    return;
-  }
-
-  if (self.destroyed) {
-    this.buffer = null;
-    return;
-  }
-
-  var availOutAfter = state[0];
-  var availInAfter = state[1];
-
-  const inDelta = handle.availInBefore - availInAfter;
-  self.bytesWritten += inDelta;
-
-  var have = handle.availOutBefore - availOutAfter;
-  if (have > 0) {
-    var out = self._outBuffer.slice(self._outOffset, self._outOffset + have);
-    self._outOffset += have;
-    self.push(out);
-  } else {
-    assert(have === 0, 'have should not go down');
-  }
-
-  if (self.destroyed) {
-    return;
-  }
-
-  // exhausted the output buffer, or used all the input create a new one.
-  if (availOutAfter === 0 || self._outOffset >= self._chunkSize) {
-    handle.availOutBefore = self._chunkSize;
-    self._outOffset = 0;
-    self._outBuffer = Buffer.allocUnsafe(self._chunkSize);
-  }
-
-  if (availOutAfter === 0) {
-    // Not actually done. Need to reprocess.
-    // Also, update the availInBefore to the availInAfter value,
-    // so that if we have to hit it a third (fourth, etc.) time,
-    // it'll have the correct byte counts.
-    handle.inOff += inDelta;
-    handle.availInBefore = availInAfter;
-
-    this.write(handle.flushFlag,
-               this.buffer, // in
-               handle.inOff, // in_off
-               handle.availInBefore, // in_len
-               self._outBuffer, // out
-               self._outOffset, // out_off
-               self._chunkSize); // out_len
-    return;
-  }
-
-  // finished with the chunk.
-  this.buffer = null;
-  this.cb();
-}
-
-function _close(engine, callback) {
-  if (callback)
-    process.nextTick(callback);
-
-  // Caller may invoke .close after a zlib error (which will null _handle).
-  if (!engine._handle)
-    return;
-
-  engine._handle.close();
-  engine._handle = null;
-}
-
 // generic zlib
 // minimal 2-byte header
 function Deflate(opts) {
@@ -645,28 +708,32 @@ function Deflate(opts) {
     return new Deflate(opts);
   Zlib.call(this, opts, DEFLATE);
 }
-inherits(Deflate, Zlib);
+ObjectSetPrototypeOf(Deflate.prototype, Zlib.prototype);
+ObjectSetPrototypeOf(Deflate, Zlib);
 
 function Inflate(opts) {
   if (!(this instanceof Inflate))
     return new Inflate(opts);
   Zlib.call(this, opts, INFLATE);
 }
-inherits(Inflate, Zlib);
+ObjectSetPrototypeOf(Inflate.prototype, Zlib.prototype);
+ObjectSetPrototypeOf(Inflate, Zlib);
 
 function Gzip(opts) {
   if (!(this instanceof Gzip))
     return new Gzip(opts);
   Zlib.call(this, opts, GZIP);
 }
-inherits(Gzip, Zlib);
+ObjectSetPrototypeOf(Gzip.prototype, Zlib.prototype);
+ObjectSetPrototypeOf(Gzip, Zlib);
 
 function Gunzip(opts) {
   if (!(this instanceof Gunzip))
     return new Gunzip(opts);
   Zlib.call(this, opts, GUNZIP);
 }
-inherits(Gunzip, Zlib);
+ObjectSetPrototypeOf(Gunzip.prototype, Zlib.prototype);
+ObjectSetPrototypeOf(Gunzip, Zlib);
 
 function DeflateRaw(opts) {
   if (opts && opts.windowBits === 8) opts.windowBits = 9;
@@ -674,38 +741,107 @@ function DeflateRaw(opts) {
     return new DeflateRaw(opts);
   Zlib.call(this, opts, DEFLATERAW);
 }
-inherits(DeflateRaw, Zlib);
+ObjectSetPrototypeOf(DeflateRaw.prototype, Zlib.prototype);
+ObjectSetPrototypeOf(DeflateRaw, Zlib);
 
 function InflateRaw(opts) {
   if (!(this instanceof InflateRaw))
     return new InflateRaw(opts);
   Zlib.call(this, opts, INFLATERAW);
 }
-inherits(InflateRaw, Zlib);
+ObjectSetPrototypeOf(InflateRaw.prototype, Zlib.prototype);
+ObjectSetPrototypeOf(InflateRaw, Zlib);
 
 function Unzip(opts) {
   if (!(this instanceof Unzip))
     return new Unzip(opts);
   Zlib.call(this, opts, UNZIP);
 }
-inherits(Unzip, Zlib);
+ObjectSetPrototypeOf(Unzip.prototype, Zlib.prototype);
+ObjectSetPrototypeOf(Unzip, Zlib);
 
 function createConvenienceMethod(ctor, sync) {
   if (sync) {
     return function syncBufferWrapper(buffer, opts) {
       return zlibBufferSync(new ctor(opts), buffer);
     };
-  } else {
-    return function asyncBufferWrapper(buffer, opts, callback) {
-      if (typeof opts === 'function') {
-        callback = opts;
-        opts = {};
-      }
-      return zlibBuffer(new ctor(opts), buffer, callback);
-    };
   }
+  return function asyncBufferWrapper(buffer, opts, callback) {
+    if (typeof opts === 'function') {
+      callback = opts;
+      opts = {};
+    }
+    return zlibBuffer(new ctor(opts), buffer, callback);
+  };
 }
 
+const kMaxBrotliParam = MathMax(...ObjectKeys(constants).map((key) => {
+  return key.startsWith('BROTLI_PARAM_') ? constants[key] : 0;
+}));
+
+const brotliInitParamsArray = new Uint32Array(kMaxBrotliParam + 1);
+
+const brotliDefaultOpts = {
+  flush: BROTLI_OPERATION_PROCESS,
+  finishFlush: BROTLI_OPERATION_FINISH,
+  fullFlush: BROTLI_OPERATION_FLUSH
+};
+function Brotli(opts, mode) {
+  assert(mode === BROTLI_DECODE || mode === BROTLI_ENCODE);
+
+  brotliInitParamsArray.fill(-1);
+  if (opts && opts.params) {
+    for (const origKey of ObjectKeys(opts.params)) {
+      const key = +origKey;
+      if (NumberIsNaN(key) || key < 0 || key > kMaxBrotliParam ||
+          (brotliInitParamsArray[key] | 0) !== -1) {
+        throw new ERR_BROTLI_INVALID_PARAM(origKey);
+      }
+
+      const value = opts.params[origKey];
+      if (typeof value !== 'number' && typeof value !== 'boolean') {
+        throw new ERR_INVALID_ARG_TYPE('options.params[key]',
+                                       'number', opts.params[origKey]);
+      }
+      brotliInitParamsArray[key] = value;
+    }
+  }
+
+  const handle = mode === BROTLI_DECODE ?
+    new binding.BrotliDecoder(mode) : new binding.BrotliEncoder(mode);
+
+  this._writeState = new Uint32Array(2);
+  // TODO(addaleax): Sometimes we generate better error codes in C++ land,
+  // e.g. ERR_BROTLI_PARAM_SET_FAILED -- it's hard to access them with
+  // the current bindings setup, though.
+  if (!handle.init(brotliInitParamsArray,
+                   this._writeState,
+                   processCallback)) {
+    throw new ERR_ZLIB_INITIALIZATION_FAILED();
+  }
+
+  ZlibBase.call(this, opts, mode, handle, brotliDefaultOpts);
+}
+ObjectSetPrototypeOf(Brotli.prototype, Zlib.prototype);
+ObjectSetPrototypeOf(Brotli, Zlib);
+
+function BrotliCompress(opts) {
+  if (!(this instanceof BrotliCompress))
+    return new BrotliCompress(opts);
+  Brotli.call(this, opts, BROTLI_ENCODE);
+}
+ObjectSetPrototypeOf(BrotliCompress.prototype, Brotli.prototype);
+ObjectSetPrototypeOf(BrotliCompress, Brotli);
+
+function BrotliDecompress(opts) {
+  if (!(this instanceof BrotliDecompress))
+    return new BrotliDecompress(opts);
+  Brotli.call(this, opts, BROTLI_DECODE);
+}
+ObjectSetPrototypeOf(BrotliDecompress.prototype, Brotli.prototype);
+ObjectSetPrototypeOf(BrotliDecompress, Brotli);
+
+
 function createProperty(ctor) {
   return {
     configurable: true,
@@ -718,7 +854,7 @@ function createProperty(ctor) {
 
 // Legacy alias on the C++ wrapper object. This is not public API, so we may
 // want to runtime-deprecate it at some point. There's no hurry, though.
-Object.defineProperty(binding.Zlib.prototype, 'jsref', {
+ObjectDefineProperty(binding.Zlib.prototype, 'jsref', {
   get() { return this[owner_symbol]; },
   set(v) { return this[owner_symbol] = v; }
 });
@@ -731,6 +867,8 @@ module.exports = {
   DeflateRaw,
   InflateRaw,
   Unzip,
+  BrotliCompress,
+  BrotliDecompress,
 
   // Convenience methods.
   // compress/decompress a string or buffer in one step.
@@ -747,10 +885,14 @@ module.exports = {
   gunzip: createConvenienceMethod(Gunzip, false),
   gunzipSync: createConvenienceMethod(Gunzip, true),
   inflateRaw: createConvenienceMethod(InflateRaw, false),
-  inflateRawSync: createConvenienceMethod(InflateRaw, true)
+  inflateRawSync: createConvenienceMethod(InflateRaw, true),
+  brotliCompress: createConvenienceMethod(BrotliCompress, false),
+  brotliCompressSync: createConvenienceMethod(BrotliCompress, true),
+  brotliDecompress: createConvenienceMethod(BrotliDecompress, false),
+  brotliDecompressSync: createConvenienceMethod(BrotliDecompress, true),
 };
 
-Object.defineProperties(module.exports, {
+ObjectDefineProperties(module.exports, {
   createDeflate: createProperty(Deflate),
   createInflate: createProperty(Inflate),
   createDeflateRaw: createProperty(DeflateRaw),
@@ -758,6 +900,8 @@ Object.defineProperties(module.exports, {
   createGzip: createProperty(Gzip),
   createGunzip: createProperty(Gunzip),
   createUnzip: createProperty(Unzip),
+  createBrotliCompress: createProperty(BrotliCompress),
+  createBrotliDecompress: createProperty(BrotliDecompress),
   constants: {
     configurable: false,
     enumerable: true,
@@ -766,16 +910,15 @@ Object.defineProperties(module.exports, {
   codes: {
     enumerable: true,
     writable: false,
-    value: Object.freeze(codes)
+    value: ObjectFreeze(codes)
   }
 });
 
 // These should be considered deprecated
 // expose all the zlib constants
-const bkeys = Object.keys(constants);
-for (var bk = 0; bk < bkeys.length; bk++) {
-  var bkey = bkeys[bk];
-  Object.defineProperty(module.exports, bkey, {
-    enumerable: true, value: constants[bkey], writable: false
+for (const bkey of ObjectKeys(constants)) {
+  if (bkey.startsWith('BROTLI')) continue;
+  ObjectDefineProperty(module.exports, bkey, {
+    enumerable: false, value: constants[bkey], writable: false
   });
 }
