@@ -3,18 +3,21 @@
 const common = require('../common');
 const assert = require('assert');
 
-const { outHeadersKey } = require('internal/http');
+const { kOutHeaders } = require('internal/http');
 const { OutgoingMessage } = require('http');
 
+const warn = 'OutgoingMessage.prototype._headers is deprecated';
+common.expectWarning('DeprecationWarning', warn, 'DEP0066');
+
 {
-  // tests for _headers get method
+  // Tests for _headers get method
   const outgoingMessage = new OutgoingMessage();
   outgoingMessage.getHeaders = common.mustCall();
   outgoingMessage._headers;
 }
 
 {
-  // tests for _headers set method
+  // Tests for _headers set method
   const outgoingMessage = new OutgoingMessage();
   outgoingMessage._headers = {
     host: 'risingstack.com',
@@ -22,7 +25,7 @@ const { OutgoingMessage } = require('http');
   };
 
   assert.deepStrictEqual(
-    Object.entries(outgoingMessage[outHeadersKey]),
+    Object.entries(outgoingMessage[kOutHeaders]),
     Object.entries({
       host: ['host', 'risingstack.com'],
       origin: ['Origin', 'localhost']
