@@ -19,13 +19,17 @@
 // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
-// Flags: --expose_internals
+// Flags: --expose-internals
 'use strict';
 const common = require('../common');
+common.skipIfDumbTerminal();
 
 const assert = require('assert');
 const readline = require('readline');
-const internalReadline = require('internal/readline');
+const {
+  getStringWidth,
+  stripVTControlCharacters
+} = require('internal/util/inspect');
 const EventEmitter = require('events').EventEmitter;
 const { Writable, Readable } = require('stream');
 
@@ -61,7 +65,7 @@ function isWarned(emitter) {
 }
 
 {
-  // set crlfDelay to float 100.5ms
+  // Set crlfDelay to float 100.5ms
   const fi = new FakeInput();
   const rli = new readline.Interface({
     input: fi,
@@ -73,7 +77,7 @@ function isWarned(emitter) {
 }
 
 {
-  // set crlfDelay to 5000ms
+  // Set crlfDelay to 5000ms
   const fi = new FakeInput();
   const rli = new readline.Interface({
     input: fi,
@@ -98,7 +102,7 @@ function isWarned(emitter) {
     rli.close();
   }
 
-  // default history size 30
+  // Default history size 30
   {
     const fi = new FakeInput();
     const rli = new readline.Interface(
@@ -126,7 +130,7 @@ function isWarned(emitter) {
     assert.ok(called);
   }
 
-  // sending a blank line
+  // Sending a blank line
   {
     const fi = new FakeInput();
     const rli = new readline.Interface(
@@ -141,7 +145,7 @@ function isWarned(emitter) {
     assert.ok(called);
   }
 
-  // sending a single character with no newline
+  // Sending a single character with no newline
   {
     const fi = new FakeInput();
     const rli = new readline.Interface(fi, {});
@@ -154,7 +158,7 @@ function isWarned(emitter) {
     rli.close();
   }
 
-  // sending a single character with no newline and then a newline
+  // Sending a single character with no newline and then a newline
   {
     const fi = new FakeInput();
     const rli = new readline.Interface(
@@ -172,7 +176,7 @@ function isWarned(emitter) {
     rli.close();
   }
 
-  // sending multiple newlines at once
+  // Sending multiple newlines at once
   {
     const fi = new FakeInput();
     const rli = new readline.Interface(
@@ -189,7 +193,7 @@ function isWarned(emitter) {
     rli.close();
   }
 
-  // sending multiple newlines at once that does not end with a new line
+  // Sending multiple newlines at once that does not end with a new line
   {
     const fi = new FakeInput();
     const rli = new readline.Interface(
@@ -206,7 +210,7 @@ function isWarned(emitter) {
     rli.close();
   }
 
-  // sending multiple newlines at once that does not end with a new(empty)
+  // Sending multiple newlines at once that does not end with a new(empty)
   // line and a `end` event
   {
     const fi = new FakeInput();
@@ -228,7 +232,7 @@ function isWarned(emitter) {
     rli.close();
   }
 
-  // sending multiple newlines at once that does not end with a new line
+  // Sending multiple newlines at once that does not end with a new line
   // and a `end` event(last line is)
 
   // \r should behave like \n when alone
@@ -288,7 +292,7 @@ function isWarned(emitter) {
     }), delay * 2);
   }
 
-  // set crlfDelay to `Infinity` is allowed
+  // Set crlfDelay to `Infinity` is allowed
   {
     const fi = new FakeInput();
     const delay = 200;
@@ -354,52 +358,52 @@ function isWarned(emitter) {
     rli.close();
   }
 
-  // constructor throws if completer is not a function or undefined
+  // Constructor throws if completer is not a function or undefined
   {
     const fi = new FakeInput();
-    common.expectsError(function() {
+    assert.throws(() => {
       readline.createInterface({
         input: fi,
         completer: 'string is not valid'
       });
     }, {
-      type: TypeError,
+      name: 'TypeError',
       code: 'ERR_INVALID_OPT_VALUE'
     });
   }
 
-  // constructor throws if historySize is not a positive number
+  // Constructor throws if historySize is not a positive number
   {
     const fi = new FakeInput();
-    common.expectsError(function() {
+    assert.throws(() => {
       readline.createInterface({
         input: fi, historySize: 'not a number'
       });
     }, {
-      type: RangeError,
+      name: 'RangeError',
       code: 'ERR_INVALID_OPT_VALUE'
     });
 
-    common.expectsError(function() {
+    assert.throws(() => {
       readline.createInterface({
         input: fi, historySize: -1
       });
     }, {
-      type: RangeError,
+      name: 'RangeError',
       code: 'ERR_INVALID_OPT_VALUE'
     });
 
-    common.expectsError(function() {
+    assert.throws(() => {
       readline.createInterface({
         input: fi, historySize: NaN
       });
     }, {
-      type: RangeError,
+      name: 'RangeError',
       code: 'ERR_INVALID_OPT_VALUE'
     });
   }
 
-  // duplicate lines are removed from history when
+  // Duplicate lines are removed from history when
   // `options.removeHistoryDuplicates` is `true`
   {
     const fi = new FakeInput();
@@ -410,6 +414,7 @@ function isWarned(emitter) {
       removeHistoryDuplicates: true
     });
     const expectedLines = ['foo', 'bar', 'baz', 'bar', 'bat', 'bat'];
+    // ['foo', 'baz', 'bar', bat'];
     let callCount = 0;
     rli.on('line', function(line) {
       assert.strictEqual(line, expectedLines[callCount]);
@@ -430,16 +435,55 @@ function isWarned(emitter) {
     assert.strictEqual(callCount, 0);
     fi.emit('keypress', '.', { name: 'down' }); // 'baz'
     assert.strictEqual(rli.line, 'baz');
+    assert.strictEqual(rli.historyIndex, 2);
     fi.emit('keypress', '.', { name: 'n', ctrl: true }); // 'bar'
     assert.strictEqual(rli.line, 'bar');
+    assert.strictEqual(rli.historyIndex, 1);
+    fi.emit('keypress', '.', { name: 'n', ctrl: true });
+    assert.strictEqual(rli.line, 'bat');
+    assert.strictEqual(rli.historyIndex, 0);
+    // Activate the substring history search.
     fi.emit('keypress', '.', { name: 'down' }); // 'bat'
     assert.strictEqual(rli.line, 'bat');
-    fi.emit('keypress', '.', { name: 'down' }); // ''
-    assert.strictEqual(rli.line, '');
+    assert.strictEqual(rli.historyIndex, -1);
+    // Deactivate substring history search.
+    fi.emit('keypress', '.', { name: 'backspace' }); // 'ba'
+    assert.strictEqual(rli.historyIndex, -1);
+    assert.strictEqual(rli.line, 'ba');
+    // Activate the substring history search.
+    fi.emit('keypress', '.', { name: 'down' }); // 'ba'
+    assert.strictEqual(rli.historyIndex, -1);
+    assert.strictEqual(rli.line, 'ba');
+    fi.emit('keypress', '.', { name: 'down' }); // 'ba'
+    assert.strictEqual(rli.historyIndex, -1);
+    assert.strictEqual(rli.line, 'ba');
+    fi.emit('keypress', '.', { name: 'up' }); // 'bat'
+    assert.strictEqual(rli.historyIndex, 0);
+    assert.strictEqual(rli.line, 'bat');
+    fi.emit('keypress', '.', { name: 'up' }); // 'bar'
+    assert.strictEqual(rli.historyIndex, 1);
+    assert.strictEqual(rli.line, 'bar');
+    fi.emit('keypress', '.', { name: 'up' }); // 'baz'
+    assert.strictEqual(rli.historyIndex, 2);
+    assert.strictEqual(rli.line, 'baz');
+    fi.emit('keypress', '.', { name: 'up' }); // 'ba'
+    assert.strictEqual(rli.historyIndex, 4);
+    assert.strictEqual(rli.line, 'ba');
+    fi.emit('keypress', '.', { name: 'up' }); // 'ba'
+    assert.strictEqual(rli.historyIndex, 4);
+    assert.strictEqual(rli.line, 'ba');
+    // Deactivate substring history search and reset history index.
+    fi.emit('keypress', '.', { name: 'right' }); // 'ba'
+    assert.strictEqual(rli.historyIndex, -1);
+    assert.strictEqual(rli.line, 'ba');
+    // Substring history search activated.
+    fi.emit('keypress', '.', { name: 'up' }); // 'ba'
+    assert.strictEqual(rli.historyIndex, 0);
+    assert.strictEqual(rli.line, 'bat');
     rli.close();
   }
 
-  // duplicate lines are not removed from history when
+  // Duplicate lines are not removed from history when
   // `options.removeHistoryDuplicates` is `false`
   {
     const fi = new FakeInput();
@@ -472,7 +516,7 @@ function isWarned(emitter) {
     rli.close();
   }
 
-  // sending a multi-byte utf8 char over multiple writes
+  // Sending a multi-byte utf8 char over multiple writes
   {
     const buf = Buffer.from('☮', 'utf8');
     const fi = new FakeInput();
@@ -501,21 +545,26 @@ function isWarned(emitter) {
       { input: fi, output: fi, terminal: true }
     );
     const keys = [];
+    const err = new Error('bad thing happened');
     fi.on('keypress', function(key) {
       keys.push(key);
       if (key === 'X') {
-        throw new Error('bad thing happened');
+        throw err;
       }
     });
-    try {
-      fi.emit('data', 'fooX');
-    } catch { }
+    assert.throws(
+      () => fi.emit('data', 'fooX'),
+      (e) => {
+        assert.strictEqual(e, err);
+        return true;
+      }
+    );
     fi.emit('data', 'bar');
     assert.strictEqual(keys.join(''), 'fooXbar');
     rli.close();
   }
 
-  // calling readline without `new`
+  // Calling readline without `new`
   {
     const fi = new FakeInput();
     const rli = readline.Interface(
@@ -531,7 +580,7 @@ function isWarned(emitter) {
     rli.close();
   }
 
-  // calling the question callback
+  // Calling the question callback
   {
     let called = false;
     const fi = new FakeInput();
@@ -570,13 +619,13 @@ function isWarned(emitter) {
       rli.question(expectedLines[0], function() {
         rli.close();
       });
-      const cursorPos = rli._getCursorPos();
+      const cursorPos = rli.getCursorPos();
       assert.strictEqual(cursorPos.rows, 0);
       assert.strictEqual(cursorPos.cols, expectedLines[0].length);
       rli.close();
     }
 
-    // sending a multi-line question
+    // Sending a multi-line question
     {
       const fi = new FakeInput();
       const rli = new readline.Interface(
@@ -586,7 +635,7 @@ function isWarned(emitter) {
       rli.question(expectedLines.join('\n'), function() {
         rli.close();
       });
-      const cursorPos = rli._getCursorPos();
+      const cursorPos = rli.getCursorPos();
       assert.strictEqual(cursorPos.rows, expectedLines.length - 1);
       assert.strictEqual(cursorPos.cols, expectedLines.slice(-1)[0].length);
       rli.close();
@@ -603,11 +652,11 @@ function isWarned(emitter) {
       });
       fi.emit('data', 'the quick brown fox');
       fi.emit('keypress', '.', { ctrl: true, name: 'a' });
-      let cursorPos = rli._getCursorPos();
+      let cursorPos = rli.getCursorPos();
       assert.strictEqual(cursorPos.rows, 0);
       assert.strictEqual(cursorPos.cols, 0);
       fi.emit('keypress', '.', { ctrl: true, name: 'e' });
-      cursorPos = rli._getCursorPos();
+      cursorPos = rli.getCursorPos();
       assert.strictEqual(cursorPos.rows, 0);
       assert.strictEqual(cursorPos.cols, 19);
       rli.close();
@@ -623,33 +672,121 @@ function isWarned(emitter) {
         terminal: terminal
       });
       fi.emit('data', 'the quick brown fox');
-      let cursorPos = rli._getCursorPos();
+      let cursorPos = rli.getCursorPos();
       assert.strictEqual(cursorPos.rows, 0);
       assert.strictEqual(cursorPos.cols, 19);
 
       // Back one character
       fi.emit('keypress', '.', { ctrl: true, name: 'b' });
-      cursorPos = rli._getCursorPos();
+      cursorPos = rli.getCursorPos();
       assert.strictEqual(cursorPos.rows, 0);
       assert.strictEqual(cursorPos.cols, 18);
       // Back one character
       fi.emit('keypress', '.', { ctrl: true, name: 'b' });
-      cursorPos = rli._getCursorPos();
+      cursorPos = rli.getCursorPos();
       assert.strictEqual(cursorPos.rows, 0);
       assert.strictEqual(cursorPos.cols, 17);
       // Forward one character
       fi.emit('keypress', '.', { ctrl: true, name: 'f' });
-      cursorPos = rli._getCursorPos();
+      cursorPos = rli.getCursorPos();
       assert.strictEqual(cursorPos.rows, 0);
       assert.strictEqual(cursorPos.cols, 18);
       // Forward one character
       fi.emit('keypress', '.', { ctrl: true, name: 'f' });
-      cursorPos = rli._getCursorPos();
+      cursorPos = rli.getCursorPos();
       assert.strictEqual(cursorPos.rows, 0);
       assert.strictEqual(cursorPos.cols, 19);
       rli.close();
     }
 
+    // Back and Forward one astral character
+    {
+      const fi = new FakeInput();
+      const rli = new readline.Interface({
+        input: fi,
+        output: fi,
+        prompt: '',
+        terminal: terminal
+      });
+      fi.emit('data', '💻');
+
+      // Move left one character/code point
+      fi.emit('keypress', '.', { name: 'left' });
+      let cursorPos = rli.getCursorPos();
+      assert.strictEqual(cursorPos.rows, 0);
+      assert.strictEqual(cursorPos.cols, 0);
+
+      // Move right one character/code point
+      fi.emit('keypress', '.', { name: 'right' });
+      cursorPos = rli.getCursorPos();
+      assert.strictEqual(cursorPos.rows, 0);
+      assert.strictEqual(cursorPos.cols, 2);
+
+      rli.on('line', common.mustCall((line) => {
+        assert.strictEqual(line, '💻');
+      }));
+      fi.emit('data', '\n');
+      rli.close();
+    }
+
+    // Two astral characters left
+    {
+      const fi = new FakeInput();
+      const rli = new readline.Interface({
+        input: fi,
+        output: fi,
+        prompt: '',
+        terminal: terminal
+      });
+      fi.emit('data', '💻');
+
+      // Move left one character/code point
+      fi.emit('keypress', '.', { name: 'left' });
+      let cursorPos = rli.getCursorPos();
+      assert.strictEqual(cursorPos.rows, 0);
+      assert.strictEqual(cursorPos.cols, 0);
+
+      fi.emit('data', '🐕');
+      cursorPos = rli.getCursorPos();
+      assert.strictEqual(cursorPos.rows, 0);
+      assert.strictEqual(cursorPos.cols, 2);
+
+      rli.on('line', common.mustCall((line) => {
+        assert.strictEqual(line, '🐕💻');
+      }));
+      fi.emit('data', '\n');
+      rli.close();
+    }
+
+    // Two astral characters right
+    {
+      const fi = new FakeInput();
+      const rli = new readline.Interface({
+        input: fi,
+        output: fi,
+        prompt: '',
+        terminal: terminal
+      });
+      fi.emit('data', '💻');
+
+      // Move left one character/code point
+      fi.emit('keypress', '.', { name: 'right' });
+      let cursorPos = rli.getCursorPos();
+      assert.strictEqual(cursorPos.rows, 0);
+      assert.strictEqual(cursorPos.cols, 2);
+
+      fi.emit('data', '🐕');
+      cursorPos = rli.getCursorPos();
+      assert.strictEqual(cursorPos.rows, 0);
+      assert.strictEqual(cursorPos.cols, 4);
+
+      rli.on('line', common.mustCall((line) => {
+        assert.strictEqual(line, '💻🐕');
+      }));
+      fi.emit('data', '\n');
+      rli.close();
+    }
+
     {
       // `wordLeft` and `wordRight`
       const fi = new FakeInput();
@@ -661,19 +798,19 @@ function isWarned(emitter) {
       });
       fi.emit('data', 'the quick brown fox');
       fi.emit('keypress', '.', { ctrl: true, name: 'left' });
-      let cursorPos = rli._getCursorPos();
+      let cursorPos = rli.getCursorPos();
       assert.strictEqual(cursorPos.rows, 0);
       assert.strictEqual(cursorPos.cols, 16);
       fi.emit('keypress', '.', { meta: true, name: 'b' });
-      cursorPos = rli._getCursorPos();
+      cursorPos = rli.getCursorPos();
       assert.strictEqual(cursorPos.rows, 0);
       assert.strictEqual(cursorPos.cols, 10);
       fi.emit('keypress', '.', { ctrl: true, name: 'right' });
-      cursorPos = rli._getCursorPos();
+      cursorPos = rli.getCursorPos();
       assert.strictEqual(cursorPos.rows, 0);
       assert.strictEqual(cursorPos.cols, 16);
       fi.emit('keypress', '.', { meta: true, name: 'f' });
-      cursorPos = rli._getCursorPos();
+      cursorPos = rli.getCursorPos();
       assert.strictEqual(cursorPos.rows, 0);
       assert.strictEqual(cursorPos.cols, 19);
       rli.close();
@@ -775,13 +912,13 @@ function isWarned(emitter) {
         terminal: terminal
       });
       fi.emit('data', 'the quick brown fox');
-      let cursorPos = rli._getCursorPos();
+      let cursorPos = rli.getCursorPos();
       assert.strictEqual(cursorPos.rows, 0);
       assert.strictEqual(cursorPos.cols, 19);
 
       // Delete left character
       fi.emit('keypress', '.', { ctrl: true, name: 'h' });
-      cursorPos = rli._getCursorPos();
+      cursorPos = rli.getCursorPos();
       assert.strictEqual(cursorPos.rows, 0);
       assert.strictEqual(cursorPos.cols, 18);
       rli.on('line', common.mustCall((line) => {
@@ -791,6 +928,31 @@ function isWarned(emitter) {
       rli.close();
     }
 
+    // deleteLeft astral character
+    {
+      const fi = new FakeInput();
+      const rli = new readline.Interface({
+        input: fi,
+        output: fi,
+        prompt: '',
+        terminal: terminal
+      });
+      fi.emit('data', '💻');
+      let cursorPos = rli.getCursorPos();
+      assert.strictEqual(cursorPos.rows, 0);
+      assert.strictEqual(cursorPos.cols, 2);
+      // Delete left character
+      fi.emit('keypress', '.', { ctrl: true, name: 'h' });
+      cursorPos = rli.getCursorPos();
+      assert.strictEqual(cursorPos.rows, 0);
+      assert.strictEqual(cursorPos.cols, 0);
+      rli.on('line', common.mustCall((line) => {
+        assert.strictEqual(line, '');
+      }));
+      fi.emit('data', '\n');
+      rli.close();
+    }
+
     // deleteRight
     {
       const fi = new FakeInput();
@@ -804,13 +966,13 @@ function isWarned(emitter) {
 
       // Go to the start of the line
       fi.emit('keypress', '.', { ctrl: true, name: 'a' });
-      let cursorPos = rli._getCursorPos();
+      let cursorPos = rli.getCursorPos();
       assert.strictEqual(cursorPos.rows, 0);
       assert.strictEqual(cursorPos.cols, 0);
 
       // Delete right character
       fi.emit('keypress', '.', { ctrl: true, name: 'd' });
-      cursorPos = rli._getCursorPos();
+      cursorPos = rli.getCursorPos();
       assert.strictEqual(cursorPos.rows, 0);
       assert.strictEqual(cursorPos.cols, 0);
       rli.on('line', common.mustCall((line) => {
@@ -820,6 +982,34 @@ function isWarned(emitter) {
       rli.close();
     }
 
+    // deleteRight astral character
+    {
+      const fi = new FakeInput();
+      const rli = new readline.Interface({
+        input: fi,
+        output: fi,
+        prompt: '',
+        terminal: terminal
+      });
+      fi.emit('data', '💻');
+
+      // Go to the start of the line
+      fi.emit('keypress', '.', { ctrl: true, name: 'a' });
+      let cursorPos = rli.getCursorPos();
+      assert.strictEqual(cursorPos.rows, 0);
+      assert.strictEqual(cursorPos.cols, 0);
+
+      // Delete right character
+      fi.emit('keypress', '.', { ctrl: true, name: 'd' });
+      cursorPos = rli.getCursorPos();
+      assert.strictEqual(cursorPos.rows, 0);
+      assert.strictEqual(cursorPos.cols, 0);
+      rli.on('line', common.mustCall((line) => {
+        assert.strictEqual(line, '');
+      }));
+      fi.emit('data', '\n');
+      rli.close();
+    }
 
     // deleteLineLeft
     {
@@ -831,13 +1021,13 @@ function isWarned(emitter) {
         terminal: terminal
       });
       fi.emit('data', 'the quick brown fox');
-      let cursorPos = rli._getCursorPos();
+      let cursorPos = rli.getCursorPos();
       assert.strictEqual(cursorPos.rows, 0);
       assert.strictEqual(cursorPos.cols, 19);
 
       // Delete from current to start of line
       fi.emit('keypress', '.', { ctrl: true, shift: true, name: 'backspace' });
-      cursorPos = rli._getCursorPos();
+      cursorPos = rli.getCursorPos();
       assert.strictEqual(cursorPos.rows, 0);
       assert.strictEqual(cursorPos.cols, 0);
       rli.on('line', common.mustCall((line) => {
@@ -860,13 +1050,13 @@ function isWarned(emitter) {
 
       // Go to the start of the line
       fi.emit('keypress', '.', { ctrl: true, name: 'a' });
-      let cursorPos = rli._getCursorPos();
+      let cursorPos = rli.getCursorPos();
       assert.strictEqual(cursorPos.rows, 0);
       assert.strictEqual(cursorPos.cols, 0);
 
       // Delete from current to end of line
       fi.emit('keypress', '.', { ctrl: true, shift: true, name: 'delete' });
-      cursorPos = rli._getCursorPos();
+      cursorPos = rli.getCursorPos();
       assert.strictEqual(cursorPos.rows, 0);
       assert.strictEqual(cursorPos.cols, 0);
       rli.on('line', common.mustCall((line) => {
@@ -876,7 +1066,7 @@ function isWarned(emitter) {
       rli.close();
     }
 
-    // multi-line cursor position
+    // Multi-line input cursor position
     {
       const fi = new FakeInput();
       const rli = new readline.Interface({
@@ -887,12 +1077,29 @@ function isWarned(emitter) {
       });
       fi.columns = 10;
       fi.emit('data', 'multi-line text');
-      const cursorPos = rli._getCursorPos();
+      const cursorPos = rli.getCursorPos();
       assert.strictEqual(cursorPos.rows, 1);
       assert.strictEqual(cursorPos.cols, 5);
       rli.close();
     }
 
+    // Multi-line prompt cursor position
+    {
+      const fi = new FakeInput();
+      const rli = new readline.Interface({
+        input: fi,
+        output: fi,
+        prompt: '\nfilledline\nwraping text\n> ',
+        terminal: terminal
+      });
+      fi.columns = 10;
+      fi.emit('data', 't');
+      const cursorPos = rli.getCursorPos();
+      assert.strictEqual(cursorPos.rows, 4);
+      assert.strictEqual(cursorPos.cols, 3);
+      rli.close();
+    }
+
     // Clear the whole screen
     {
       const fi = new FakeInput();
@@ -905,7 +1112,7 @@ function isWarned(emitter) {
       const lines = ['line 1', 'line 2', 'line 3'];
       fi.emit('data', lines.join('\n'));
       fi.emit('keypress', '.', { ctrl: true, name: 'l' });
-      const cursorPos = rli._getCursorPos();
+      const cursorPos = rli.getCursorPos();
       assert.strictEqual(cursorPos.rows, 0);
       assert.strictEqual(cursorPos.cols, 6);
       rli.on('line', common.mustCall((line) => {
@@ -916,52 +1123,45 @@ function isWarned(emitter) {
     }
   }
 
-  // isFullWidthCodePoint() should return false for non-numeric values
-  [true, false, null, undefined, {}, [], 'あ'].forEach((v) => {
-    assert.strictEqual(internalReadline.isFullWidthCodePoint('あ'), false);
-  });
+  // Wide characters should be treated as two columns.
+  assert.strictEqual(getStringWidth('a'), 1);
+  assert.strictEqual(getStringWidth('あ'), 2);
+  assert.strictEqual(getStringWidth('谢'), 2);
+  assert.strictEqual(getStringWidth('고'), 2);
+  assert.strictEqual(getStringWidth(String.fromCodePoint(0x1f251)), 2);
+  assert.strictEqual(getStringWidth('abcde'), 5);
+  assert.strictEqual(getStringWidth('古池や'), 6);
+  assert.strictEqual(getStringWidth('ノード.js'), 9);
+  assert.strictEqual(getStringWidth('你好'), 4);
+  assert.strictEqual(getStringWidth('안녕하세요'), 10);
+  assert.strictEqual(getStringWidth('A\ud83c\ude00BC'), 5);
+  assert.strictEqual(getStringWidth('👨‍👩‍👦‍👦'), 8);
+  assert.strictEqual(getStringWidth('🐕𐐷あ💻😀'), 9);
+  // TODO(BridgeAR): This should have a width of 4.
+  assert.strictEqual(getStringWidth('⓬⓪'), 2);
+  assert.strictEqual(getStringWidth('\u0301\u200D\u200E'), 0);
 
-  // wide characters should be treated as two columns.
-  assert.strictEqual(internalReadline.isFullWidthCodePoint('a'.charCodeAt(0)),
-                     false);
-  assert.strictEqual(internalReadline.isFullWidthCodePoint('あ'.charCodeAt(0)),
-                     true);
-  assert.strictEqual(internalReadline.isFullWidthCodePoint('谢'.charCodeAt(0)),
-                     true);
-  assert.strictEqual(internalReadline.isFullWidthCodePoint('고'.charCodeAt(0)),
-                     true);
-  assert.strictEqual(internalReadline.isFullWidthCodePoint(0x1f251), true);
-  assert.strictEqual(internalReadline.getStringWidth('abcde'), 5);
-  assert.strictEqual(internalReadline.getStringWidth('古池や'), 6);
-  assert.strictEqual(internalReadline.getStringWidth('ノード.js'), 9);
-  assert.strictEqual(internalReadline.getStringWidth('你好'), 4);
-  assert.strictEqual(internalReadline.getStringWidth('안녕하세요'), 10);
-  assert.strictEqual(internalReadline.getStringWidth('A\ud83c\ude00BC'), 5);
-
-  // check if vt control chars are stripped
+  // Check if vt control chars are stripped
   assert.strictEqual(
-    internalReadline.stripVTControlCharacters('\u001b[31m> \u001b[39m'),
+    stripVTControlCharacters('\u001b[31m> \u001b[39m'),
     '> '
   );
   assert.strictEqual(
-    internalReadline.stripVTControlCharacters('\u001b[31m> \u001b[39m> '),
+    stripVTControlCharacters('\u001b[31m> \u001b[39m> '),
     '> > '
   );
   assert.strictEqual(
-    internalReadline.stripVTControlCharacters('\u001b[31m\u001b[39m'),
+    stripVTControlCharacters('\u001b[31m\u001b[39m'),
     ''
   );
   assert.strictEqual(
-    internalReadline.stripVTControlCharacters('> '),
+    stripVTControlCharacters('> '),
     '> '
   );
-  assert.strictEqual(internalReadline
-    .getStringWidth('\u001b[31m> \u001b[39m'), 2);
-  assert.strictEqual(internalReadline
-    .getStringWidth('\u001b[31m> \u001b[39m> '), 4);
-  assert.strictEqual(internalReadline
-    .getStringWidth('\u001b[31m\u001b[39m'), 0);
-  assert.strictEqual(internalReadline.getStringWidth('> '), 2);
+  assert.strictEqual(getStringWidth('\u001b[31m> \u001b[39m'), 2);
+  assert.strictEqual(getStringWidth('\u001b[31m> \u001b[39m> '), 4);
+  assert.strictEqual(getStringWidth('\u001b[31m\u001b[39m'), 0);
+  assert.strictEqual(getStringWidth('> '), 2);
 
   {
     const fi = new FakeInput();
@@ -979,7 +1179,7 @@ function isWarned(emitter) {
     assert.strictEqual(isWarned(process.stdout._events), false);
   }
 
-  // can create a new readline Interface with a null output argument
+  // Can create a new readline Interface with a null output argument
   {
     const fi = new FakeInput();
     const rli = new readline.Interface(
@@ -1037,7 +1237,7 @@ function isWarned(emitter) {
 const crlfDelay = Infinity;
 
 [ true, false ].forEach(function(terminal) {
-  // sending multiple newlines at once that does not end with a new line
+  // Sending multiple newlines at once that does not end with a new line
   // and a `end` event(last line is)
 
   // \r\n should emit one line event, not two
@@ -1106,3 +1306,26 @@ const crlfDelay = Infinity;
     }), delay);
   }
 });
+
+// Ensure that the _wordLeft method works even for large input
+{
+  const input = new Readable({
+    read() {
+      this.push('\x1B[1;5D'); // CTRL + Left
+      this.push(null);
+    },
+  });
+  const output = new Writable({
+    write: common.mustCall((data, encoding, cb) => {
+      assert.strictEqual(rl.cursor, rl.line.length - 1);
+      cb();
+    }),
+  });
+  const rl = new readline.createInterface({
+    input: input,
+    output: output,
+    terminal: true,
+  });
+  rl.line = `a${' '.repeat(1e6)}a`;
+  rl.cursor = rl.line.length;
+}
